#include <stdio.h>
#include <string.h>
//#include <cuda.h>
//#include <cutil.h>
#include "utility.cuh"

#define VERSION "0.1"
#define VALUETOWRITE 1234
#define MEMCHUNKS 19

__global__ void write_value(int* d_mem, long unsigned int n_ints,
        unsigned int nx, unsigned int ny)
{
    // 3d thread index
    unsigned int x = blockDim.x * blockIdx.x + threadIdx.x;
    unsigned int y = blockDim.y * blockIdx.y + threadIdx.y;
    unsigned int z = blockDim.z * blockIdx.z + threadIdx.z;

    // 1d thread index
    long unsigned int idx = x + nx*y + nx*ny*z;

    if (idx < n_ints) {
        __syncthreads();
        d_mem[idx] = VALUETOWRITE;
    }
}

int main(int argc, char** argv)
{
    int target_device = 0;

    if (argc == 2 &&
            (strcmp(argv[1], "-v") == 0 || strcmp(argv[1], "--version") == 0)) {
        printf("CUDA memory scrubber, version %s\n", VERSION);
        printf("License GPLv3+: GNU GPL version 3 or later "
                "http://gnu.org/licenses/gpl.html\n"
                "There is NO WARRANTY, to the extent permitted by law.\n"
                "Written by Anders Damsgaard <andersd@riseup.net>\n"
                "Maintained at "
                "https://github.com/anders-dc/cuda-memscrub\n");
        exit(0);
    } else if (argc == 3 &&
            (strcmp(argv[1], "-d") == 0 || strcmp(argv[1], "--device") == 0)) {
        target_device = atoi(argv[2]);
    } else if (argc == 2 &&
            (strcmp(argv[1], "-h") == 0 || strcmp(argv[1], "--help") == 0)) {
        printf("CUDA memory scrubber. Usage:\n"
                " %s [OPTIONS]\n", argv[0]);
        printf("Options:\n"
                "\t-h, --help\t\tshow this information\n"
                "\t-v, --version\t\tshow version information\n"
                "\t-d <n>, --device <n>\tscrub device with index n\n");
        exit(0);
    } else if (argc > 1) {
        fprintf(stderr, "argument not understood. See %s for usage "
                "information\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    checkForCudaErrors("Before initializing CUDA device");

    int device_count;
    cudaGetDeviceCount(&device_count);
    cudaDeviceProp prop;

    if (device_count == 0) {
        fprintf(stderr, "Error: No CUDA-enabled devices available. Bye.\n");
        exit(EXIT_FAILURE);
    } else if (target_device >= device_count) {
        fprintf(stderr, "Error: No CUDA-enabled device by id %d is detected.\n",
                target_device);
        exit(EXIT_FAILURE);
    } else {
        cudaGetDeviceProperties(&prop, target_device);
        printf("target: device %d, %s\n", target_device, prop.name);
        cudaChooseDevice(&target_device, &prop);
    }

    size_t mem_size = prop.totalGlobalMem;
    printf("global memory size: %lu bytes\n", mem_size);

    long unsigned int n_ints = mem_size/sizeof(int)/(MEMCHUNKS+1);
    printf("overwriting the first %ld bytes, corresponding to %ld int values "
            "or the first %.1f%% of the global device memory.\n",
            n_ints*sizeof(int)*MEMCHUNKS, n_ints*MEMCHUNKS,
            (float)100*n_ints*sizeof(int)*MEMCHUNKS/prop.totalGlobalMem);

    int* d[MEMCHUNKS];  // array of device pointers
    int i;
    for (i=0; i<MEMCHUNKS; i++) {
        if (cudaMalloc((void**)&d[i], n_ints*sizeof(int))
                    == cudaErrorMemoryAllocation) {
            fprintf(stderr, "Error: Could not allocate the requested amount of "
                    "global memory on the device.\n");
            cudaDeviceReset();
            exit(EXIT_FAILURE);
        }
    }
    checkForCudaErrors("After memory allocation");

    dim3 dimBlock(prop.maxThreadsPerBlock, 1, 1);
    unsigned int grid_size = iDivUp(n_ints, prop.maxThreadsPerBlock);
    if (grid_size > prop.maxGridSize[0]) {
        fprintf(stderr, "Error: The device cannot handle a grid large enough to"
                " handle the array size\n");
        cudaDeviceReset();
        exit(EXIT_FAILURE);
    }
    dim3 dimGrid(grid_size, 1, 1);

    for (i=0; i<MEMCHUNKS; i++) {
        write_value<<<dimGrid, dimBlock>>>(d[i], n_ints, 1, 1);
        cudaThreadSynchronize();
        checkForCudaErrors("After write_value", i);
    }


    for (i=0; i<MEMCHUNKS; i++) {
        cudaFree(d[i]);
        checkForCudaErrors("After cudaFree(d[i])", i);
    }

    cudaDeviceReset();

    return 0;
}
